/*							       	    HTFWrite.c
**	FILE WRITER
**
**	(c) COPYRIGHT CERN 1994.
**	Please first read the full copyright statement in the file COPYRIGH.
**
**	This version of the stream object just writes to a C file.
**	The file is assumed open and left open.
**
**	Bugs:
**		strings written must be less than buffer size.
**
**      History:
**         HFN: wrote it
**         HWL: converted the caching scheme to be hierachical by taking
**              AL code from Deamon
**
*/

/* Library include files */
#include "tcp.h"
#include "HTUtils.h"
#include "HTString.h"
#include "HTFormat.h"
#include "HTAlert.h"
#include "HTList.h"
#include "HTFWrite.h"
#include "RoverBase.h"

#define CACHE_LIMIT 100		/* files */

#ifdef Mips
#define L_tmpnam 64
#endif




#include "HTParse.h"





/*		Stream Object
**		------------
*/

struct _HTStream {
	CONST HTStreamClass *	isa;
	
	FILE *			fp;
	BOOL			leave_open;     /* Close file? HFN 08/02-94 */
	char * 			end_command;
	BOOL 			remove_on_close;
	BOOL			announce;
	char *			filename;
	HTRequest *		request;	/* saved for callback */
	BOOL (*callback) PARAMS((struct _HTRequest * req, void * filename));
	CacheItem *		cache;
};


/* ------------------------------------------------------------------------- */
/*  			     BASIC STREAM CLASSES			     */
/* ------------------------------------------------------------------------- */

/*
**
**		B L A C K    H O L E    C L A S S
**
**	There is only one black hole instance shared by anyone
**	who wants a black hole.  These black holes don't radiate,
**	they just absorb data.
*/
PRIVATE void HTBlackHole_put_character ARGS2(HTStream *, me, char, c)
{
    return;
}

PRIVATE void HTBlackHole_put_string ARGS2(HTStream *, me, CONST char*, s)
{
    return;
}

PRIVATE void HTBlackHole_write ARGS3(HTStream *, me, CONST char*, s, int, l)
{
    return;
}

PRIVATE int HTBlackHole_free ARGS1(HTStream *, me)
{
    return 0;
}

PRIVATE int HTBlackHole_abort ARGS2(HTStream *, me, HTError, e)
{
    return EOF;
}


/*	Black Hole stream
**	-----------------
*/
PRIVATE CONST HTStreamClass HTBlackHoleClass =
{		
	"BlackHole",
	HTBlackHole_free,
	HTBlackHole_abort,
	HTBlackHole_put_character, 	HTBlackHole_put_string,
	HTBlackHole_write
}; 

PRIVATE HTStream HTBlackHoleInstance =
{
	&HTBlackHoleClass,

	NULL,
	NO,				/* HENRIK 08/02-94 */
	NULL,
	NO,
	NO,
	NULL,
	NULL,
	NULL,
	NULL,
};

/* Black hole creation */
PUBLIC HTStream * HTBlackHole NOARGS
{
    if (TRACE)
	fprintf(TDEST, "BlackHole... Created\n");
    return &HTBlackHoleInstance;
}


/*	HTThroughLine
**	-------------
**
** This function is a dummy function that returns the same output stream
** as given as a parameter. Henrik 01/03-94
*/
PUBLIC HTStream* HTThroughLine ARGS5(
	HTRequest *,		request,
	void *,			param,
	HTFormat,		input_format,
	HTFormat,		output_format,
	HTStream *,		output_stream)	            /* Only one used */
{
    return output_stream;
}







