 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTLog.h
 *
 * Description: Rover pending message log
 *
 * Last Modified: 10-Feb-95 (adj)
 *
 * Global Functions:
 *    PUBLIC HTStream *HTAppRun(HTRequest *request,
 *				void *param,
 *				HTFormat input_format,
 *			        HTFormat output_format,
 *				HTStream *output_stream)
 *        Converter for application/x-application MIME stream. Downloads
 *        application binary code, creates pipes for IPC, and forks off
 *        a separate process.
 *
 * Global Variables: PUBLIC CONST HTStreamClass HTApplication
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef HTLOG_H
#define HTLOG_H

extern int NetAvail;
#define MAX_NET_PRIORITY 	4
#define MIN_NET_PRIORITY 	0
#define MAX_RETRIES		10

typedef struct _HTNetQ HTNetQ;
struct _HTNetQ {
  int logID;		/* Entry identifier */
  int dirty;		/* Entry needs to be written to stable storage */
  int local;		/* Server is localhost? */
  int smtp;		/* Send request via SMTP? */
  char *url;		/* Target url */
  int session;		/* Session ID */
  char *value;		/* Additional data to be sent */
  char *command;	/* Command to send */
  RoverCBList callback;	/* Server callback ID */
  QRPCState *state;	/* State associated with the queue entry */
  HTRequest *req;	/* Request associated with the queue entry */
  int retry;		/* Number of retry attempts */
  int msgid;		/* Message ID for the request */
  HTNetQ *next;		/* Next queue entry */
  int priority;		/* Priority of the request */
  struct timeval QTv;	/* When the request was enqueued */
};

#define ROVER_LOG_VERSION 2

extern void LogSave(int force);
extern RoverErr LogInit();
extern void LogDeleteFile();
extern void LogSetRequest(HTRequest *req, QRPCState *state,  char *url,
			  RoverCBList callback);
extern int LogClose();
extern int LogFlush();
extern int LogDelete(int LogID, int batched);
extern int LogClose();
extern int LogEnqueue(int session, char *url, char *value, char *command,
		      RoverCBList callback, QRPCState *state, int priority);
extern int LogEnqueueSMTP(int session, char *url, char *value, int size,
			  int priority);
extern int LogEnqueueSMTP_TCP(int session, char *url, char *value, int size,
			      int priority);
extern int LogRetry(HTStream *me);
extern HTNetQ *LogFindLogID(int logID);

extern HTNetQ *RoverLogQ[MAX_NET_PRIORITY+2];
extern HTNetQ *RoverSentQ;
extern int LogChanged;
extern char *RoverLogDir;
extern int RoverHardFlush;
extern int RoverUseSMTP;
extern int RoverNoLog;
extern int RoverDelayedFlush;

#endif /* HTLOG_H */
