 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverHTTPUtil.c
 *
 * Description: HTTP utility functions
 *
 *
 ***********************************************************************
 *   Copyright 1997
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR
#define ROVER_HTTP_PRIVATE

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTCache.h"
#include "RoverZlib.h"
#include "html.h"
#include "RoverTCP.h"
#include "HTRoverHTTP.h"
#include "RoverFile.h"
#include "RoverEscape.h"

extern void makePrefetch(char *url, char *sourceurl, char *sourcetitle);

PUBLIC int RoverHTTPPrefetch=1;	/* levels of recursion for prefetching */

PRIVATE int HTTPSendHdr = 1;	/* Send MIME header with first request */
PRIVATE char *HTTPAgent = NULL;	/* Browser agent MIME line */
PRIVATE char *HTTPAccept =NULL;	/* Browser accept MIME line */
PRIVATE time_t time1;		/* time of most recent request */
PRIVATE long preftime=3600;	/* seconds-- if mail gets to server
				   faster than this, don't prefetch. */

PRIVATE char HTTPbuffer[1024], from[1024], status[32], MIMEHdr[2048];
PRIVATE char *TargetWindow = "target=\"RoverWWW\"";
PRIVATE RoverBuffer *RoverCacheBuffer = NULL;

PRIVATE char *RoverHTTPErrString =
"HTTP/1.0 200 OK\nContent-Type: text/html

<html>
<head>
<meta http-equiv=\"Pragma\" content=\"no-cache\">
</head>
<title>Rover Browser Proxy Error</title>
<h1>Rover Browser Proxy Home Error</h1>
An internal Rover Browser Proxy Error occurred while processing your request.
Please resubmit your request again.<p>

If the error continues to occur, contact
<a href=\"mailto:adj@pdos.lcs.mit.edu\">adj@pdos.lcs.mit.edu</A>
</html>";

PRIVATE char *HomePage =
"HTTP/1.0 200 OK\nContent-Type: text/html

<html>
<head>
<meta http-equiv=\"Pragma\" content=\"no-cache\">
</head>
<title>Rover Browser Proxy home page</title>
<h1>Rover Browser Proxy Home Page</h1>
Welcome to the Rover Browser Proxy Home Page. To follow a link, click on it.
An entry will appear in the cache window, marked (pending). When the
entry changes to show the View, Reload, and Delete buttons, click on
View to see the document, or click on the original link again. To
delete the cache entry, click on Delete. To Reload the document, click
on Reload. Here are a few interesting links:<p>

The <a href=\"http://home/cache\">Rover Browser Proxy Cache</a> page.<p>
The <a href=\"http://budapest.lcs.mit.edu/cgi-bin/demo/cgi-client-new.gcc\">AutoFocus Query Engine</a>.<p>
The <a href=\"http://www.w3.org/hypertext/WWW/TheProject.html\">World Wide Web </a> home page.<p>
The <a href=\"http://home.netscape.com/\">Netscape</a> home page.<p>
The <a href=\"http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/NCSAMosaicHome.html\">NCSA Mosaic </a> home page.<p>
The <a href=\"http://www.rover.lcs.mit.edu/\">Rover</a> home page.<p>
<a href=\"http://www.pdos.lcs.mit.edu/~adj/\">Anthony</a>'s home page.<p>
<a href=\"http://www.psrg.lcs.mit.edu/~aldel/\">Alan</a>'s home page.<p>
<hr><address>
<a href=\"http://www.rover.lcs.mit.edu/\">The Rover Toolkit</a><p>
Anthony Joseph and Alan deLespinasse<br>
<a href=\"mailto:adj@pdos.lcs.mit.edu\">{adj,aldel}@pdos.lcs.mit.edu</A>
</address>
</html>";

PRIVATE char *TestPage =
"HTTP/1.0 200 OK\nContent-Type: text/html

<HTML>
<HEAD>
<TITLE>Thesis Test Page</TITLE>
</HEAD>
<BODY>

<H1>Thesis Test Page</H1>

<P>Here are a few interesting links:</P>

<P> 1. The <A HREF=\"http://spot.lcs.mit.edu/ns/www/TheProject.html\">World
Wide Web </A>home page.</P>

<P> 2. The <A HREF=\"http://spot.lcs.mit.edu/ns/netscape/index.html\">Netscape</A> home page.</P>

<P> 3. The <A HREF=\"http://spot.lcs.mit.edu/ns/mosaic/NCSAMosaicHome.html\">NCSA
Mosaic </A>home page.</P>

<P> 4. The <A HREF=\"http://spot.lcs.mit.edu/ns/rover/index.html\">Rover</A> home page.</P>

<P> 5. <A HREF=\"http://spot.lcs.mit.edu/ns/adj/index.html\">Anthony</A>'s home page.</P>

<P> 6. <A HREF=\"http://spot.lcs.mit.edu/ns/aldel/index.html\">Alan</A>'s home page.</P>

<P> 7. <A HREF=\"http://spot.lcs.mit.edu/ns/kaashoek/index.html\">Professor Frans
Kaashoek</A>'s home page.</P>

<P> 8. The <A HREF=\"http://spot.lcs.mit.edu/ns/pdos/index.html\">Parallel and Distributed
Operating Systems</A> group home page.</P>

<P> 9. The <A HREF=\"http://spot.lcs.mit.edu/ns/psg/index.html\">Parallel Software Group</A>
home page.</P>

<P> 10. The <A HREF=\"http://spot.lcs.mit.edu/ns/lcs/index.html\">Laboratory for Computer Science</A>'s
home page.</P>

<P> </P>

</BODY>
</HTML>";

PUBLIC char *HttpText(HTTPState state)
{
  switch (state) {
  case HTTP_BEGIN:
    return("HTTP_BEGIN");
    break;
  case HTTP_NEED_REQUEST:
    return("HTTP_NEED_REQUEST");
    break;
  case HTTP_GOT_REQUEST:
    return("HTTP_GOT_REQUEST");
    break;
  case HTTP_SERVICE_CACHE:
    return("HTTP_SERVICE_CACHE");
    break;
  case HTTP_GET_CACHELINES:
    return("HTTP_GET_CACHELINES");
    break;
  case HTTP_PROCESS_CACHELINES:
    return("HTTP_PROCESS_CACHELINES");
    break;
  case HTTP_PROCESS_CACHE:
    return("HTTP_PROCESS_CACHE");
    break;
  case HTTP_READ_MIMEHDR:
    return("HTTP_READ_MIMEHDR");
    break;
  case HTTP_SAVE_MIMEHDR:
    return("HTTP_SAVE_MIMEHDR");
    break;
  case HTTP_READ_BODY:
    return("HTTP_READ_BODY");
    break;
  case HTTP_SAVE_BODY:
    return("HTTP_SAVE_BODY");
    break;
  case HTTP_GOT_BODY:
    return("HTTP_GOT_BODY");
    break;
  case HTTP_SERVICE_SERVER:
    return("HTTP_SERVICE_SERVER");
    break;
  case HTTP_READ_SERVERBODY:
    return("HTTP_READ_SERVERBODY");
    break;
  case HTTP_SAVE_SERVERBODY:
    return("HTTP_SAVE_SERVERBODY");
    break;
  case HTTP_GOT_SERVERBODY:
    return("HTTP_GOT_SERVERBODY");
    break;
  case HTTP_WRITE_CLIENT:
    return("HTTP_WRITE_CLIENT");
    break;
  case HTTP_SEND_MAIL:
    return("HTTP_SEND_MAIL");
    break;
  case HTTP_GET_MAILREQUEST:
    return("HTTP_GET_MAILREQUEST");
    break;
  case HTTP_GOT_MAILREQUEST:
    return("HTTP_GOT_MAILREQUEST");
    break;
  case HTTP_GET_MAILBODY:
    return("HTTP_GET_MAILBODY");
    break;
  case HTTP_GOT_MAILBODY:
    return("HTTP_GOT_MAILBODY");
    break;
  case HTTP_UPDATE_INLINES:
    return("HTTP_UPDATE_INLINES");
    break;
  case HTTP_WRITE_CLIENTHDR:
    return("HTTP_WRITE_CLIENTHDR");
    break;
  case HTTP_CHECK_HDR:
    return("HTTP_CHECK_HDR");
    break;
  case HTTP_WRITE_CLIENTTYPE:
    return("HTTP_WRITE_CLIENTTYPE");
    break;
  case HTTP_WRITE_CLIENTPAGE:
    return("HTTP_WRITE_CLIENTPAGE");
    break;
  case HTTP_FINISH_PAGE:
    return("HTTP_FINISH_PAGE");
    break;
  case HTTP_SERVICE_CLIENT:
    return("HTTP_SERVICE_CLIENT");
    break;
  case HTTP_SERVICE_ROVER_CLIENT:
    return("HTTP_SERVICE_ROVER_CLIENT");
    break;
  case HTTP_CLEANUP:
    return("HTTP_CLEANUP");
    break;
  case HTTP_ABORT:
    return("HTTP_ABORT");
    break;
  default:
    return("<unknown>");
    break;
  }
}

/* Create a new cache file name for an HTTP document and return a FILE
 *   pointer to the file. Also updates the cache entry, if it exists.
 */
PUBLIC FILE *MakeNewPath(Http *http)
{
  char *path;
  FILE *file;

  if (!(path = malloc(strlen("/RoverHTTPXXXXXXXX")+strlen(RoverCacheDir)+1))) {
    fprintf(stderr,"Out of memory! [MakeNewPath:RoverHTTPUtil.c]\n");
    return(NULL);
  }
  sprintf(path, "%s/RoverHTTPXXXXXXXX", RoverCacheDir);
  if ((file = RoverGetUniqFile(path)) == NULL) {
    fprintf(stderr, "File %s create error: %s [MakeNewPath:RoverHTTPUtil.c]\n",
	    path, strerror(errno));
    return(NULL);
  }

  if (!http->p && !(http->p = RoverHTTPGetPageInfo(http->url))) {
    fprintf(stderr,"New cache entry %s missing [MakeNewPath:RoverHTTPUtil].\n",
	    http->url);
    return(NULL);
  }
  http->p->path = path;
  http->path = strdup(path);

  if (RoverDebug)
    fprintf(stderr, "Path is: %s [1] [MakeNewPath:RoverHTTPUtil.c]\n", path);
  return(file);
}

/* Returns 1 if the page information exists for url. */
PUBLIC int RoverHTTPPageInfoExists(char *url)
{
  int size;

  return(RoverCacheSearchData(url, HTTPSession, &size) != NULL);
}

/* Returns the page information for a url if it exists. */
PUBLIC RoverHTTP *RoverHTTPGetPageInfo(char *url)
{
  char *data;
  int size;

  if ((data = RoverCacheSearchData(url, HTTPSession, &size)) == NULL) {
     if (RoverDebug) fprintf(stderr, "Page %s not found\n", url);
    return(NULL);
  } else
    return(UnmarshallHTTP(&data));
}

/* Deletes a cache entry. */
PRIVATE inline void cachedelete(RoverHTTP *p)
{
  if (RoverDebug) fprintf(stderr, "Deleting page `%s'\n", p->url);
  RoverCacheFlush(p->url, HTTPSession);
  if (p->path) unlink(p->path);
  FreeHTTP(p);
  RoverUpdateHTTPAll();

  /* Delete the cache display, as it is stale */
  if (RoverCacheBuffer) {
    RoverBufferFree(RoverCacheBuffer);
    RoverCacheBuffer = NULL;
  }
}

/* Deletes a cache entry, specified by the "page" field. */
PUBLIC int cachedeletepage(char *url)
{
  RoverHTTP *p = RoverHTTPGetPageInfo(url);

  if (!p) {
    if (RoverDebug) fprintf(stderr, "Page %s not found while deleting\n", url);
    return(-1);
  } else {
    cachedelete(p);
    return(0);
  }
}

/* Returns 1 if the entry is not visible and is not referenced, directly
   or indirectly, by any visible entry. A is assumed to reference B iff
   B->source points to A. */
PRIVATE int cachegarbage(RoverHTTP *p)
{
  RoverHTTP *q;
  int rc;

  if (p->visible) return 0;
  if (!strcmp(p->source, HomeURL)) return 0;
  if (RoverDebug) fprintf(stderr, "GC checking source of %s (%s)\n", p->url,
			  p->source);
  if ((q = RoverHTTPGetPageInfo(p->source)) == NULL) {
    if (RoverDebug) fprintf(stderr, "Source %s not found --> garbage\n",
			    p->source);
    return(1);
  } else {
    if (RoverDebug) fprintf(stderr, "Source %s found --> not garbage\n",
			    p->source);
    rc = cachegarbage(q);
    FreeHTTP(q);
    if (RoverDebug) fprintf(stderr, "Ancestors %sfound\n", rc ? "not ": "");
    return(rc);
  }
}

/* Garbage collector. Deletes any records that are "garbage" according to
   the above function. */
PUBLIC void cachegc()
{
  RoverHTTP *p;
  Tcl_HashSearch *search = NULL;
  char *data;
  int size;
  time_t objTime, lastRef;

  /* 
     1. First mark all pages with 0 gc counts
     2. For each page, increment the gc count for each referenced page
   */

restart:
  while ((data = RoverCacheSearchByType(ROVER_HTTP, &size, &objTime, &lastRef,
					&search)) != NULL) {
    p = UnmarshallHTTP(&data);
    if (RoverDebug) HTTPPrint(p);
    if (cachegarbage(p)) {
      cachedelete(p);
      if (search) free(search);
      search = NULL;
      goto restart;	/* Restart gc (no other choice, since Tcl hash tables
			   don't support deletes during searches :-( ) */
    }
    FreeHTTP(p);
  }
}

PUBLIC void RoverUpdateHTTPAll()
{
  if (RoverDebug) fprintf(stderr, "Sending http cache updates\n");
  RoverAppDoUpdate(ROVER_WATCH_HTTP, RoverUpdateHTTP);
}

PRIVATE inline char *RoverHTTPGetSourceTitle(char *url)
{
  char *title = NULL;
  RoverHTTP *q;

  if (!url || !strlen(url))
    return(NULL);
  
  if (!strcmp(url, HomeURL))
    title = strdup(HomeTitle);
  else if ((q = RoverHTTPGetPageInfo(url)) != NULL) {
    if (q->pagetitle && strlen(q->pagetitle)) title = strdup(q->pagetitle);
    free(q);
  }
  return(title);
}


PRIVATE void RoverHTTPFixSource(char *url, char *source)
{
  char *title = NULL;
  RoverHTTP *p;
  RoverErr rc;

  if (!(title = RoverHTTPGetSourceTitle(source))) return;
  if (!(p = RoverHTTPGetPageInfo(url))) return;

  if (RoverDebug) fprintf(stderr,"Found sourcetitle %s for %s\n",title,source);
  if (p->sourcetitle) free(p->sourcetitle);
  if (p->source) free(p->source);
  p->sourcetitle = strdup(title);
  p->source = strdup(source);
  rc = HTTPUpdate(p);
  if (rc != ROVER_OK)
    fprintf(stderr,
	    "HTTP cache update failes [RoverHTTPFixSource:RoverHTTPUtil.c]\n");
  FreeHTTP(p);
}

struct _RoverHTML {
  int uid;			/* Id of tcl object (== load/request order)*/
  char *html;			/* HTML */
};
typedef struct _RoverHTML RoverHTML;

int comp(const void *a, const void *b)
{
  RoverHTML *html1 = (RoverHTML *)a, *html2 = (RoverHTML *)b;
  return(html2->uid - html1->uid);
}

PRIVATE char *RoverHTTPMakeHTMLEntry(RoverHTTP *p, time_t t)
{
  struct tm* tim;
  char ds[32];

  if (!p->visible || p->image) return("");
  tim = localtime(&t);
  sprintf(ds, "%02d/%02d/%02d %02d:%02d%s  ", tim->tm_mon+1, tim->tm_mday,
	  tim->tm_year % 100,(tim->tm_hour+11) % 12 +1, tim->tm_min,
	  (tim->tm_hour>=12) ? "pm" : "am");

  if (p->source && strlen(p->source))
    sprintf(from, "<a href=\"%s\" %s>%s</a>", p->source, TargetWindow,
	    (strlen(p->sourcetitle) ? p->sourcetitle : p->source));
  else 
    sprintf(from, "<i>unspecified</i>");
  
  if (!p->arrived) sprintf(status, "pending");
  else sprintf(status, "arrived / %s", (p->viewed ? "viewed" : "not viewed"));

  sprintf(HTTPbuffer,"%s <a href=\"%s\" %s>%s</a> <br>
requested from %s (%s)<p>\n", ds, p->url, TargetWindow,
	  (strlen(p->pagetitle) ? p->pagetitle : p->url), from, status);
  return(strdup(HTTPbuffer));
}

PUBLIC RoverBuffer *RoverHTTPMakeHTML()
{
  RoverHTML *html;
  char *data;
  int size, i = 0, count = 0;
  Tcl_HashSearch *search = NULL;
  RoverHTTP *p;
  time_t objTime, lastRef;

  if (RoverDebug) fprintf(stderr, "Enter RoverMakeHTML\n");
  if (RoverCacheBuffer) return(RoverCacheBuffer);

  if (RoverDebug) fprintf(stderr, "RoverMakeHTML: recreate buffer\n");
  RoverCacheBuffer = RoverBufferInit(1024);
  while ((data = RoverCacheSearchByType(ROVER_HTTP, &size, &objTime, &lastRef,
					&search)) != NULL) {
    p = UnmarshallHTTP(&data);
    if (RoverDebug) { 
      fprintf(stderr, ">> p->url %s p->visible %d p->image %d\n", p->url,
	      p->visible, p->image);
      fprintf(stderr, ">> p->pagetitle %s p->source %s p->sourcetitle %s\n\n",
	      p->pagetitle, p->source, p->sourcetitle);
    }
    if (p->visible && !p->image) count++;
    FreeHTTP(p);
  }

  if ((html = (RoverHTML *) malloc(count * sizeof(RoverHTML))) == NULL) {
    fprintf(stderr,"Out of memory! [RoverUpdate:HTRover.c]\n");
    RoverBufferFree(RoverCacheBuffer);
    RoverCacheBuffer = NULL;
    return(NULL);
  }

  while ((data = RoverCacheSearchByType(ROVER_HTTP, &size, &objTime, &lastRef,
					&search)) != NULL) {
    p = UnmarshallHTTP(&data);
    if (p->visible && !p->image) {
      html[i].uid = p->uid;
      html[i++].html = RoverHTTPMakeHTMLEntry(p,(objTime ? objTime : lastRef));
    }
    FreeHTTP(p);
  }
  qsort(html, count, sizeof(RoverHTML), comp);

  RoverBufferAppend(RoverCacheBuffer,
"HTTP/1.0 200 OK\nContent-Type: text/html\n

<html>
<meta http-equiv=\"Refresh\" content=\"5\">
<head>
<meta http-equiv=\"Pragma\" content=\"no-cache\">
</head>
<title>Rover Browser Proxy Cache</title>
<h1>Rover Browser Proxy Cache</h1>
<dl>");
  if (count) {
    for (i = 0; i < count; i++) {
      RoverBufferAppend(RoverCacheBuffer, "<dt> ");
      RoverBufferAppend(RoverCacheBuffer, html[i].html);
      RoverBufferAppend(RoverCacheBuffer, "</dl>\n");
      free(html[i].html);
    }
    RoverBufferAppend(RoverCacheBuffer, "</dl>\n");
  } else
    RoverBufferAppend(RoverCacheBuffer,"<p>No documents in proxy cache.<p>\n");
  RoverBufferAppend(RoverCacheBuffer, "<hr><address>
<a href=\"http://www.rover.lcs.mit.edu/\">The Rover Toolkit</a><p>
Anthony Joseph and Alan deLespinasse<br>
<a href=\"mailto:adj@pdos.lcs.mit.edu\">{adj,aldel}@pdos.lcs.mit.edu</A>
</address></html>");
  return(RoverCacheBuffer);
}

PUBLIC void RoverUpdateHTTP(int appID, int toPipe, int watch)
{
  char *rbuf, *rbuf2, *data;
  int csize, count = 0, size = 0;
  Tcl_HashSearch *search = NULL;
  RoverHTTP *p;
  time_t objTime, lastRef;
  RoverErr rv;

  /* 
     The marshalled http info has the following format:
        [# http records]
	 [<Page load time>
	  <Marshalled record>]
	          :
		  :
		  :
   */

  while ((data = RoverCacheSearchByType(ROVER_HTTP, &csize, &objTime, &lastRef,
					&search)) != NULL) {
    p = UnmarshallHTTP(&data);
    if (RoverDebug) { 
      fprintf(stderr, ">> p->url %s p->visible %d p->image %d\n", p->url,
	      p->visible, p->image);
      fprintf(stderr, ">> p->pagetitle %s p->source %s p->sourcetitle %s\n\n",
	      p->pagetitle, p->source, p->sourcetitle);
    }
    if (p->visible && !p->image) {
      count++;
      size += (csize+(2*sizeof(int)));
    }
    FreeHTTP(p);
  }
  size += sizeof(int);		/* One int for total size */
  if ((rbuf2 = rbuf = malloc(size)) == NULL) {
    fprintf(stderr,"Out of memory! [RoverUpdate:HTRover.c]\n");
    return;
  }

  if (RoverDebug) fprintf(stderr, "Sending %d http records\n", count);
  MarshallInt(&rbuf2, count);
  while ((data = RoverCacheSearchByType(ROVER_HTTP, &csize, &objTime, &lastRef,
					&search)) != NULL) {
    p = UnmarshallHTTP(&data);
#if 0
    HTTPPrint(p);
#endif
    if (p->visible && !p->image) {
      MarshallInt(&rbuf2, (objTime ? objTime : lastRef));
      MarshallBin(&rbuf2, csize, data);
    }
    FreeHTTP(p);
  }
  ASSERT(size>=(rbuf2-rbuf));
  rv = RoverServerSend(toPipe, 0, ROVER_HTTPCACHE_UPDATE, size, rbuf, 1);
  if (rv != ROVER_OK)
    fprintf(stderr, "%s: Send fail %s [RoverUpdateHTTP:RoverHTTPUtil.c]\n",
	    myAppID, RoverErrType(rv));

}

/* Plain text callback for HTML parser. Sets the pagetitle field if
   it's between a <title> and a </title>, as determined by the tagCallback
   function. */
void textCallback(HTMLparser *p)
{
  Http *http = (Http *)p->data;
  char *t;

  if ((!http->intitle) || (http->p->pagetitle && strlen(http->p->pagetitle)))
    return;
  t = &(p->buf[strspn(p->buf, " \t\r\n")]);
  if (!*t) return;
  t = strtok(t, "\n\r");
  if (!t || !*t) return;
  http->p->pagetitle = strdup(t);
}

/* Called whenever an HTML tag is found. Keeps a count of the number
   of inlined images and hrefs found. */
PUBLIC void tagCallback(HTMLparser *p)
{
  URLinfo uinfo;
  char *es, *t = strtok(p->buf," =<>\t\n\r");
  Http *http = (Http *)p->data;

  if (RoverDebug) fprintf(stderr, "Tag callback: %s\n", t);
  if (!strcasecmp(t, "a")) {
    for (;;) {
      t = strtok(NULL, " <>=\t\n\r");	/* find the HREF reference */
      if (!t) return;
      if (!strcasecmp(t, "href")) continue;

      t = strtok(NULL, " <>=\"\t\n\r");	/* find the URL */
      if (!t) return;
      if (RoverDebug) 
	fprintf(stderr, "Found href URL: %s\n  [Base URL: %s]\n", t,
		http->p->url);
      if (!strncasecmp("mailto:", t, 7) || !strncasecmp("news:", t, 5) ||
	  !strncasecmp("javascript:", t, 7)) break; /* Ignore some hrefs */

      http->p->hrefs=(char **)realloc(http->p->hrefs, 
				      sizeof(char *)*(http->p->hrefCount+1));
      if (http->p->hrefs == NULL) {
	fprintf(stderr,"Out of memory [2]! [tagCallback:RoverHTTPUtil.c]\n");
	RoverExit(-1);
      }
      uinfo.url = copyString(t);
      uinfo.base = http->p->url;
      if ((es = convertURL(&uinfo)) != NULL) {
	if (RoverDebug) 
	  fprintf(stderr, "Ignoring failed URL conversion: %s\n", es);
	free(uinfo.url);
      }
      http->p->hrefs[http->p->hrefCount++] = strdup(uinfo.url);
      if (RoverDebug) {
	fprintf(stderr, "Absolute href URL: %s\n", uinfo.url);
	fprintf(stderr, "Prefetch is %d\n", http->p->prefetch);
      }
      if (http->p->prefetch > 0)	/* Add a prefetch entry, if needed */
	makePrefetch(uinfo.url, http->p->url, http->p->pagetitle);
      free(uinfo.url);
      break;
    }
  } else if (!strcasecmp(t, "title")) {
    http->intitle = 1;
  } else if (!strcasecmp(t, "/title")) {
    http->intitle = 0;
  } else if (strcasecmp(t, "img") && strcasecmp(t, "image") &&
	     strcasecmp(t, "fig") && strcasecmp(t, "body"))
     return;
  /* Everything else is an image */

  /* First, locate the URL for the image */
  for (;;) {
    t = strtok(NULL, " <>=\t\n\r");	/* find the SRC or BACKGROUND ref */
    if (!t) return;
    if (strcasecmp(t, "src") && strcasecmp(t, "background")) continue;

    t = strtok(NULL, " <>=\"\t\n\r");	/* find the URL */
    if (!t) return;
    if (RoverDebug) 
      fprintf(stderr, "Found inlined URL: %s\n  [Base URL: %s]\n", t,
	      http->p->url);
    http->p->inlines = (char **)realloc(http->p->inlines, 
					sizeof(char *) *
					(http->p->inlineCount+1));
    if (http->p->inlines == NULL) {
      fprintf(stderr,"Out of memory [2]! [tagCallback:RoverHTTPUtil.c]\n");
      RoverExit(-1);
    }
    uinfo.url = copyString(t);
    uinfo.base = http->p->url;
    if ((es = convertURL(&uinfo)) != NULL) {
      if (RoverDebug) 
	fprintf(stderr, "Ignoring failed URL conversion: %s\n", es);
      free(uinfo.url);
    }
    http->p->inlines[http->p->inlineCount++] = strdup(uinfo.url);
    if (RoverDebug) {
      fprintf(stderr, "Absolute inlined URL: %s\n", uinfo.url);
      fprintf(stderr, "Prefetch is %d\n", http->p->prefetch);
    }
    if (http->p->prefetch >= 0)	/* Add a prefetch entry, if needed */
      makePrefetch(uinfo.url, http->p->url, http->p->pagetitle);
    free(uinfo.url);
    break;
  }
}

PUBLIC RoverErr HTTPUpdate(RoverHTTP *p)
{
  int size = HTTPMarshallSize(p);
  char *data, *ptr;
  RoverErr rc;

  if ((ptr = data = malloc(size)) == NULL) {
    fprintf(stderr, "Out of memory [HTTPUpdate:RoverHTTPUtil.c]\n");
    return(ROVER_NO_MEM);
  }
  MarshallHTTP(&data, p);
  rc = RoverCacheUpdate(p->url, HTTPSession, ROVER_HTTP,ROVER_BEST_EFFORT,NULL,
		   (time_t) 0,(time_t) 0,(time_t)time(NULL), size, ptr,NULL,0);

  /* Delete the cache display, as it is stale */
  if (RoverCacheBuffer) {
    RoverBufferFree(RoverCacheBuffer);
    RoverCacheBuffer = NULL;
  }
  if (rc != ROVER_OK)
    fprintf(stderr, "Cache insert %d::%s fail: %d [HTTPUpdate:RoverHTTPUtil.c]\n",
	    HTTPSession, p->url, rc);
  return(rc);
}

PUBLIC RoverErr HTTPInsert(RoverHTTP *p, time_t objTime, time_t delay,
			   time_t last)
{
  int size = HTTPMarshallSize(p);
  char *data, *ptr;
  RoverErr rc;

  if ((ptr = data = malloc(size)) == NULL) {
    fprintf(stderr, "Out of memory [HTTPInsert:RoverHTTPUtil.c]\n");
    return(ROVER_NO_MEM);
  }
  MarshallHTTP(&data, p);
  rc = RoverCacheInsertData(p->url, HTTPSession, ROVER_HTTP, ROVER_BEST_EFFORT,
			    ptr, size, objTime, delay, last, NULL, 0);

  /* Delete the cache display, as it is stale */
  if (RoverCacheBuffer) {
    RoverBufferFree(RoverCacheBuffer);
    RoverCacheBuffer = NULL;
  }
  if (rc != ROVER_OK)
    fprintf(stderr, "Cache insert fail %d::%s [HTTPInsert:RoverHTTPUtil.c]\n",
	    HTTPSession, p->url);
  return(rc);
}


/* Tell the browser client that its request was valid, but yielded no
   information. This returns control to the client quickly. */
PUBLIC void punt(Http *http, HTTPState nextState)
{
  if (RoverDebug) fprintf(stderr, "Punting (\"no content\" response).\n");
  http->tcp->ptr = "HTTP/1.0 204 nothing\n\n";
  http->tcp->size = strlen(http->tcp->ptr);
  http->tcp->offset = 0;
  http->nextState = nextState;
  http->state = HTTP_WRITE_CLIENT;
}

PUBLIC void handlehome(Http *http)
{
  RoverBuffer *buf;
  if (!strcasecmp(http->url, HomeURL)) {
    if (RoverDebug) fprintf(stderr, "Just requesting the home page.\n");
    http->tcp->ptr = HomePage;
    http->tcp->size = strlen(HomePage);
  } else if (!strncasecmp(http->url+strlen(HomeURL),"cache",strlen("cache"))) {
    buf = RoverHTTPMakeHTML();
    if (!buf) {
      http->tcp->ptr = RoverHTTPErrString;
      http->tcp->size = strlen(RoverHTTPErrString);
    } else {      
      http->tcp->ptr = (char *)RoverBufferBuf(buf);
      http->tcp->size = RoverBufferSize(buf);
    }
  } else if (!strncasecmp(http->url+strlen(HomeURL), "test", strlen("test"))) {
    if (RoverDebug) fprintf(stderr, "Just requesting the test page.\n");
    http->tcp->ptr = TestPage;
    http->tcp->size = strlen(TestPage);
  }
  http->tcp->offset = 0;
  http->nextState = HTTP_CLEANUP;
  http->state = HTTP_WRITE_CLIENT;
}

/* Make new cache entry */
PRIVATE RoverHTTP *RoverHTTPNewEntry(char *url, char *source,
				     char *sourcetitle, int visible, int query)
{
  RoverHTTP *p = (RoverHTTP *)malloc(sizeof(RoverHTTP));
  p->pagetitle = p->path = NULL;
  p->source = source;
  p->sourcetitle = sourcetitle;

  p->prefetch = p->inlineCount = p->hrefCount = p->msgID = 0;
  p->arrived = p->viewed = p->image = 0;
  p->url = strdup(url);
  p->visible = visible;
  p->uid = ++counter;
  p->type = NULL;
  p->inlines = p->hrefs = NULL;
  return(p);
}


/* Generates a message to send to the access manager. The access
 * manager will then send a TCP or SMTP based QRPC to the server,
 * requesting a document.
 */
PUBLIC void mailsetup(Http *http)
{
  FILE *f;
  char *source = NULL, *sourcetitle = NULL, *data;
  int doinline = 1;
  static char buffer[BUFSIZE];
  RoverErr rc;

  if (RoverDebug)
    fprintf(stderr, "Generating HTTP request [mailsetup:RoverHTTPUtil.c]\n");

  http->p = NULL;
  if (RoverHTTPPageInfoExists(http->url)) {
    fprintf(stderr, "Found cache entry for %s during mailsetup!\n", http->url);
    http->state = HTTP_ABORT;
    return;
  }
  
  http->p = RoverHTTPNewEntry(http->url, source, sourcetitle, 1, http->query);

  /* Create a placeholder cache file */
  if (!(f = MakeNewPath(http))) {
    fprintf(stderr, "File create failed [mailsetup:RoverHTTPUtil.c]\n");
    http->state = HTTP_ABORT;
  }
  fclose(f);

  rc=HTTPInsert(http->p, (time_t) time(NULL), (time_t) 0, (time_t) time(NULL));
  if (rc != ROVER_OK) {
    fprintf(stderr, "HTTP cache insert failed\n");
    http->state = HTTP_ABORT;
    return;
  }
  FreeHTTP(http->p);
  http->p = NULL;
  if (RoverDebug) fprintf(stderr, "New cache entry created.\n");

  http->mail = RoverBufferInit(1024);
  data = RoverEscape(http->url, URL_XPALPHAS);
  sprintf(buffer, "X-url: %s\n", data);
  if (data != http->url) free(data);
  RoverBufferAppend(http->mail, buffer);
  sprintf(buffer, "X-inline: %s\n", doinline ? "yes" : "no");
  RoverBufferAppend(http->mail, buffer);
  sprintf(buffer, "X-path: %s\n", http->path);
  RoverBufferAppend(http->mail, buffer);
  time1 = time(NULL);
  sprintf(buffer, "X-time1: %lu\n", (unsigned long) time1);
  RoverBufferAppend(http->mail, buffer);
  sprintf(buffer, "X-prefetch: %d\n", RoverHTTPPrefetch);
  RoverBufferAppend(http->mail, buffer);
  sprintf(buffer, "X-preftime: %ld\n", preftime);
  RoverBufferAppend(http->mail, buffer);
  sprintf(buffer, "X-compress: %d\n", RoverCompress);
  RoverBufferAppend(http->mail, buffer);
  data = strtok(http->reqline, "\n\r");
  sprintf(buffer, "X-request: %s\n", data);	/* send first line */
  RoverBufferAppend(http->mail, buffer);

  if (HTTPSendHdr) MIMEHdr[0] = '\0';
  http->tcp->offset = 0;
  http->state = HTTP_GET_MAILREQUEST;/* Get rest of original request, if any */
  http->tcp->size = BUFSIZE;
}

/* Create an empty entry for a page being prefetched. */
PUBLIC void makePrefetch(char *url, char *sourceurl, char *sourcetitle)
{
  RoverHTTP *p;
  RoverErr rc;

  if (RoverDebug)
    fprintf(stderr,"Prefetch: %s [makePrefetch:RoverHTTPUtil.c]\n",url);

  /* Make sure the entry doesn't already exist */
  if (RoverHTTPPageInfoExists(url)) return;
  p = RoverHTTPNewEntry(url, strdup(sourceurl), strdup(sourcetitle), 1, 0);

  rc = HTTPInsert(p, (time_t) 0, (time_t) 0, (time_t) 0);
  if (rc != ROVER_OK) {
    fprintf(stderr, "HTTP cache insert failed\n");
    return;
  }

  if (RoverDebug) fprintf(stderr, "New cache entry created.\n");
  FreeHTTP(p);
}

/* Sends a cached document to sock. If it's in HTML, parses it to count
   inlined images. */
PUBLIC void sendpage(Http *http)
{
  if (RoverDebug)
    fprintf(stderr,"Sending %s\nOpening `%s'\n", http->p->url, http->p->path);
  if (!(http->file = fopen(http->p->path,"r"))) {
    fprintf(stderr, "Error opening cache file %s\n", http->p->path);
    free(http->p->path);
    http->p->path = NULL;
    http->state = HTTP_ABORT;
    return;
  }
  fcntl(fileno(http->file), F_SETFD, 1);	/* Set to close on exec */
  if (!strncasecmp(http->p->url,"http:",5)) 	/* only HTTP sends MIME hdrs */
    http->state = HTTP_WRITE_CLIENTHDR;
  else
    http->state = HTTP_WRITE_CLIENTTYPE;
}

PUBLIC void HTTPProcessMailReq(Http *http)
{
  char *buf, *t;
  int ignore = 0, rc = strspn(http->tcp->line," \t\r\n");

/*  fprintf(stderr, "rc %d, strlen %d\n", rc, strlen(http->tcp->line)); */
  if (rc == strlen(http->tcp->line)) {
    /*
     * End of headers
     *  If there's a Content-Length, we need to send that much more data.
     */
    if (RoverDebug) fprintf(stderr, "End of HTTP headers\n");
    strcat(MIMEHdr, http->tcp->line);
    if (HTTPSendHdr) {
      RoverBufferAppend(http->mail, "X-MIMEHdr: ");
      t = RoverEscape(MIMEHdr, URL_XPALPHAS);
      buf = RoverEscape(t, URL_XPALPHAS);
      if (t != MIMEHdr) free(t);
      fprintf(stderr, "2x encoded MIME hdr is: %s\n", buf);
      RoverBufferAppend(http->mail, buf);
      if (buf != t) free(buf);
    }
    HTTPSendHdr = http->tcp->offset = 0;
    http->state = HTTP_GET_MAILBODY;
    return;
  }

  /* Otherwise, process the MIME lines */

  if (!strncasecmp(http->tcp->line, "content-length:", 15))
    sscanf(&(http->tcp->line[15]),"%d", &(http->tcp->size));
  else if (!strncasecmp(http->tcp->line, "if-modified-since:", 7)) {
    ignore = 1;		/* Ignore line (for now) */
  } else if (!strncasecmp(http->tcp->line, "user-agent:", 11)) {
    if (!HTTPAgent) 
      HTTPAgent = strdup(&(http->tcp->line[11]));
    else if (strcmp(HTTPAgent, &(http->tcp->line[11]))) {
	fprintf(stderr,
		"Warning! Browser agent (%s) doesn't match expected (%s)\n",
		HTTPAgent, &(http->tcp->line[11]));
	free(HTTPAgent);
	HTTPAgent = strdup(&(http->tcp->line[11]));
	HTTPSendHdr = 1;
      }
  } else if (!strncasecmp(http->tcp->line, "accept:", 7)) {
    if (!HTTPAccept) 
      HTTPAccept = strdup(&(http->tcp->line[11]));
    else if (strcmp(HTTPAccept, &(http->tcp->line[11]))) {
      fprintf(stderr,
	      "Warning! Browser accept (%s) doesn't match expected (%s)\n",
	      HTTPAccept, &(http->tcp->line[11]));
      free(HTTPAccept);
      HTTPAccept = strdup(&(http->tcp->line[11]));
      HTTPSendHdr = 1;
    }
  } else if (!strncasecmp(http->tcp->line, "referer:", 8)) {
    buf = strdup(http->tcp->line);
    t = strtok(&(buf[8]), " \t\n\r");
    if (RoverDebug) fprintf(stderr, "Referer: %s\n", t);
    RoverHTTPFixSource(http->url, t);
    free(buf);
    
    ignore = 1;		/* Ignore line */
  } else if (!strncasecmp(http->tcp->line, "host:", 5) ||
	     !strncasecmp(http->tcp->line, "proxy-connection:", 17) ||
	     !strncasecmp(http->tcp->line, "cookie:", 7)) {
    ignore = 1;		/* Ignore lines */
    if (RoverDebug) fprintf(stderr, "%s: Ignoring MIME line\n", myAppID);
  }

  if (!ignore && HTTPSendHdr) strcat(MIMEHdr, http->tcp->line);
  http->tcp->offset = 0;
  http->state = HTTP_GET_MAILREQUEST;
}

PUBLIC void HTTPServiceClient(Http *http)
{
  if (!strncmp(http->url, HomeURL, strlen(HomeURL))) {
    handlehome(http);
    return;
  }

  /* We consider it a query if the POST method is used or there are data
     fields after a ? in the URL. */
  if (!strcasecmp(http->method,"POST") || strchr(http->url,'?')) {
    http->query = 1;
    if (RoverDebug) fprintf(stderr, "It's a query.\n");
  } else {
    http->query = 0;
    if (RoverDebug) fprintf(stderr, "It's not a query.\n");
    if (RoverDebug) fprintf(stderr, "Just your basic GET.\n");
  }

  if (!(http->p = RoverHTTPGetPageInfo(http->url))) {
    if (RoverDebug) fprintf(stderr, "Cache miss.\n");
    punt(http, HTTP_SEND_MAIL);
  } else {
    if (http->p->arrived) {
      if (RoverDebug) fprintf(stderr, "Cache hit.\n");
      if (!http->p->visible)			/* it was prefetched */
	http->update = http->p->visible = 1;
      sendpage(http);
    } else 		/* It's already on its way. */
      punt(http, HTTP_UPDATE_INLINES);
  }
}




#ifdef OBSOLETE
/* All of the code below this line is now obsolete */


/* Given a MIME type and a URL, returns 1 if it thinks browser will sprout an
   external viewer. Either arg can be NULL, but not both. */
PUBLIC int extviewer(char *type, char *url)
{
  if (type) {
    if (!strncasecmp(type,"text/",5) ||	!strcasecmp(type,"application/html") ||
        !strcasecmp(type,"application/x-html") ||
	!strcasecmp(type,"application/x-wais-source") ||
	!strcasecmp(type,"application/x-hdf") ||
	!strcasecmp(type,"application/x-netcdf") ||
	!strcasecmp(type,"application/hdf") ||
	!strcasecmp(type,"application/netcdf"))
      return 0;
    return 1;
  } else {
    char *ext = strrchr(url,'.');
    if (!ext) return 0;
    if (!strcasecmp(ext,".gif") || !strcasecmp(ext,".jpg") ||
	!strcasecmp(ext,".xbm") || !strcasecmp(ext,".mpg") ||
	!strcasecmp(ext,".au") || !strcasecmp(ext,".aiff") ||
	!strcasecmp(ext,".aifc") || !strcasecmp(ext,".snd") ||
	!strcasecmp(ext,".tar") || !strcasecmp(ext,".uu") || 
	!strcasecmp(ext,".saveme") || !strcasecmp(ext,".dump") || 
	!strcasecmp(ext,".tif") || !strcasecmp(ext,".bin") || 
	!strcasecmp(ext,".tiff") || !strcasecmp(ext,".jpeg") || 
	!strcasecmp(ext,".mpeg") || !strcasecmp(ext,".ps") ||
	!strcasecmp(ext,".rgb") || !strcasecmp(ext,".rtf") || 
	!strcasecmp(ext,".pdf") || !strcasecmp(ext,".mime") ||
	!strcasecmp(ext,".ram") || !strcasecmp(ext,".ra") ||
	!strcasecmp(ext,".dvi") || !strcasecmp(ext,".wav"))
      return 1;
  }
  return 0;
}

/* Called whenever an HTML tag is found. Simply keeps a count of the number
   of inlined images found. */
PRIVATE void browserTagCallback(HTMLparser *p)
{
  char *t = strtok(p->buf," =<>\t\n\r");
  if (strcasecmp(t, "img") && strcasecmp(t, "image") &&
	     strcasecmp(t, "fig") && strcasecmp(t, "body"))
     return;
}

/* Gets and processes a buffer containing multiple newline separated fields.
   Invokes the fcn on each field (including the newline). Handles the
   return code from function appropriately. */
PRIVATE void NLGetLines(Http *http, NLFunction fcn, HTTPState nextState)
{
  int rc, size = (BUFSIZE - http->tcp->size);

  if (size <= 0) {
    fprintf(stderr, "Out of buffer space [HTTPGetBuffer:RoverHTTPUtil.c]\n");
    http->state = HTTP_ABORT;
    return;
  }
  rc = readbuffer(http->tcp->sock, http->tcp->line + http->tcp->offset, size);
  switch (rc) {
  case -1:		/* EOF */
    http->state = nextState;
    break;
  case 0:		/* No more data for now, keep going */
    break;
  case -2:		/* An error occurred (other than EINTR or EAGAIN) */
    fprintf(stderr, "Readbuffer error [HTTPGetBuffer:RoverHTTPUtil.c]\n");
    http->state = HTTP_ABORT;
    break;
  default:		/* Read some number of bytes */
    http->tcp->size += rc;
    rc = NLProcessLine(http, fcn, http->tcp->line, http->tcp->size, nextState);
    if (rc == -1) return;
    http->tcp->size -= rc;
    http->tcp->offset = http->tcp->size;
    break;
  }
}
#endif

