/*         		    					     HTBrowse.c
**	HYPERTEXT BROWSER FOR DUMB TERMINALS
**
**	(c) COPYRIGHT CERN 1994.
**	Please first read the full copyright statement in the file COPYRIGH.
**
** Compilation-time macro options
**
**	VL		Version number, quoted eg "1.2a"
*/

#define ROVER_ACCESSMGR

#include "RoverBase.h"
#include <ctype.h>
#include "HTUtils.h"	/* WWW general purpose macros */
#include "HTBrowse.h"	/* Things exported, short names */
#if 0
#include "GridText.h"	/* Hypertext definition */

#include "HTFormat.h"
#endif
#include "HTTCP.h"	/* TCP/IP utilities */
#include "HTAnchor.h"   /* Anchor class */
#include "HTParse.h"    /* WWW address manipulation */
#include "HTAccess.h"   /* WWW document access network code */
#if 0
#include "HTHistory.h"	/* Navigational aids */
#include "HTML.h"	/* For parser */
#endif
#include "HTError.h"
#include "HTAlert.h"
#include "HTTP.h"
#include "RoverEvent.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "HTLog.h"
#include "HTCache.h"
#include "RoverTest.h"

/* Public Variables etc. */
PUBLIC char *HTAppName = "Rover Access Mgr";	/* Application name */
PUBLIC char *HTAppVersion = VL;			/* Application version */
PUBLIC int errno;
 
/* Type definitions and global variables local to this module */
PRIVATE char **AppArgv = NULL;	/* Application argv from command line */
PRIVATE char *tracefile_root = NULL;  	    	/* Root of trace file name */
PRIVATE char *HTTraceFileName = NULL;  	    	/* Root of trace file name */
PRIVATE FILE *HTTracefile = NULL;		/* File for trace output */

/* ------------------------------------------------------------------------- */
/*				HELP FUNCTIONS				     */
/* ------------------------------------------------------------------------- */

/*									 ErrMsg
**
**   General Error Message for Rover Server. If verbose mode is on, then
**   the message is passed to stderr.
**
*/
PRIVATE void ErrMsg(char *Msg, char *Str)
{
  fprintf(stderr, "Warning: %s (%s)\n", Msg, Str ? Str : "");
}

/* ------------------------------------------------------------------------- */
/*				  MAIN PROGRAM				     */
/* ------------------------------------------------------------------------- */

PRIVATE void usage(char *name)
{
  fprintf(stderr,
	  "
Usage: %s\t [-C #] [-F dir] [-P #] [-S #] [-Z] [-c #]
\t\t [-b #] [-f] [-l] [-n] [-p #]
\t\t [-?] [-help] [-d #] [-t name] [-v trace]
\t\t [-H #] [-N name] [-Q #] [-s #] [-m #] [-q #] [-R string] [-U url]
\t\t \[<url>] [-- [args ...]]\n",
	  name);

fprintf(stderr, "
url \t\t Specify the URL of the initially loaded client application.
-- [args ...]\t Specify arguments for Rover application.

-C 0/1\t\t Disable/enable automatic checking of network status.
-F dir\t\t Specify the Access Manager's logging directory.
-P prefetch\t\t Specify the HTTP prefetch depth.
-S 0/1\t\t Disable/enable the use of SMTP instead of TCP.
-Z\t\t Reset all session message identifiers.
-c level\t\t Specify the zlib compression level.
-b 0/1\t\t Disable/enable batching of QRPCs.
-f\t\t Enable synchronous flushing of QRPC log.
-l limit\t Specify the maximum in-core cache size.
-n\t\t Disable logging of QRPCs to stable storage.
-p port\t\t Specify the Access Manager's HTTP port.

Diagnostic:
-?\t\t Show this information.
-help\t\t Show this information.
-d level\t Specify the debugging level.
-t name\t\t Specify the trace file for HT events.
-v trace\t Enable tracing of HT events [disabled].

QRPC and HTTP timing tests:
-H count\t Specify the number of HTTP tests to be run.
-N name\t\t Specify the name used for timing tests.
-Q size\t\t Specify the size of each QRPC.
-s secs\t\t Specify delay (in seconds) after issuing each QRPC.
-m 0/1\t\t Disable/enable multiple outstanding QRPC requests.
-q count\t Specify the number of QRPCs to issue.
-R string\t Reset string to send to the server before running tests.
-U url\t URL to use for HTTP tests.
-r\t\t Measure time to prepare QRPCs for transmission.
-e\t\t Measure time to finish processing QRPCs.
\n");
  exit(-1);
}

int main ARGS2(int, argc, char **, argv)
{
  int return_status = 0;	
  int ignore = 0;
  int arg;			       /* Argument number as we scan */
  BOOL tracefile_flag = NO;
  BOOL keyword = 0;
#ifdef ultrix
  char *url = "/binaries/base-dec.app";
#elif defined(linux)
  char *url = "/binaries/base-linux.app";
#elif defined(sparc)
  char *url = "/binaries/base-sun.app";
#else
  char *url = NULL;
#endif
  int HTid;
  QRPCState *state;
  
  /* Check for command line options */
  for (arg=1; arg<argc ; arg++) {
    if (!ignore && (*argv[arg] == '-')) {
      
      /* - alone => error */
      if (argv[arg][1] == 0) {
	ErrMsg("Bad Command Line Argument", argv[arg]);
	
	/* --: ignore all else (as args); treat as keywords */
      } else if (!strcmp(argv[arg], "--")) {
	ignore = 1;
	


	/* Rover reset message ids */
      } else if (!strcmp(argv[arg], "-Z")) {
	RoverResetMsgIDs = 1;

	/* Measure time to finish QRPC flag */
      } else if (!strcmp(argv[arg], "-e")) {
	RoverTestFinish = 1;

	/* Rover logging on/off flag */
      } else if (!strcmp(argv[arg], "-n")) {
	RoverNoLog = 1;

	/* Rover HTTP prefetch depth */
      } else if (!strcmp(argv[arg], "-P")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverHTTPPrefetch = atoi(argv[++arg]);

	/* Rover HTTP test URL */
      } else if (!strcmp(argv[arg], "-U")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverHTTPurl = argv[++arg];

	/* Rover HTTP test count */
      } else if (!strcmp(argv[arg], "-H")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverHTTPTestNum = atoi(argv[++arg]);

	/* Rover log directory */
      } else if (!strcmp(argv[arg], "-F")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverLogDir = argv[++arg];

	/* Rover in-core memory size limit */
      } else if (!strcmp(argv[arg], "-l")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverCacheMemLimit = atoi(argv[++arg]);

	/* Rover compression level */
      } else if (!strcmp(argv[arg], "-c")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverCompress = atoi(argv[++arg]);

	/* Rover batching */
      } else if (!strcmp(argv[arg], "-b")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverBatch = atoi(argv[++arg]);

	/* Rover httpd port */
      } else if (!strcmp(argv[arg], "-p")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverPort = atoi(argv[++arg]);

	/* Rover check network status */
      } else if (!strcmp(argv[arg], "-C")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverCheckNet = atoi(argv[++arg]);

	/* -? or -help: show the command line help page */
      } else if (!strcmp(argv[arg], "-?") ||
		 !strcmp(argv[arg], "-help")) {
	usage(argv[0]);
	exit(0);
	
	/* Debug flag */
      } else if (!strcmp(argv[arg], "-d")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverDebug = atoi(argv[++arg]);

	/* Test name */
      } else if (!strcmp(argv[arg], "-N")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverTestName = argv[++arg];

	/* QRPC/TCP timing sleep value */
      } else if (!strcmp(argv[arg], "-s")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverSleepTime = atoi(argv[++arg]);

	/* QRPC/TCP timing test flag */
      } else if (!strcmp(argv[arg], "-q")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverQRPCTcpTest = atoi(argv[++arg]);

	/* QRPC/TCP timing test size flag */
      } else if (!strcmp(argv[arg], "-Q")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverQRPCTcpSize = atoi(argv[++arg]);

	/* Timing test reset string flag */
      } else if (!strcmp(argv[arg], "-R")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverResetString = argv[++arg];

	/* QRPC/TCP timing test multiple outstanding msgs flag */
      } else if (!strcmp(argv[arg], "-m")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverQRPCTcpTestMulti = atoi(argv[++arg]);

	/* Hard log flush flag */
      } else if (!strcmp(argv[arg], "-f")) {
	  RoverHardFlush = 1;

	/* SMTP or TCP flag */
      } else if (!strcmp(argv[arg], "-S")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  RoverUseSMTP = atoi(argv[++arg]);

	/* QRPC prep measurement flag */
      } else if (!strcmp(argv[arg], "-r")) {
	RoverTestPrep = 1;

	/* Trace file */
      } else if (!strcmp(argv[arg], "-t")) {
	if (arg+1 < argc && *argv[arg+1] != '-')
	  HTTraceFileName = argv[++arg];
	tracefile_flag = YES;
	/* Verify: Turns on trace */
      } else if (!strncmp(argv[arg], "-v", 2)) {
#if 0
	char *p = argv[arg]+2;
#endif
	ROVER_TraceFlag = 0;
#if 0
	for(; *p; p++) {
	  switch (*p) {
	  case 'a': 	ROVER_TraceFlag += SHOW_ANCHOR_TRACE; break;
	  case 'p':	ROVER_TraceFlag += SHOW_PROTOCOL_TRACE; break;
	  case 's':	ROVER_TraceFlag += SHOW_SGML_TRACE; break;
	  case 't':	ROVER_TraceFlag += SHOW_THREAD_TRACE; break;
	  case 'u': 	ROVER_TraceFlag += SHOW_URI_TRACE; break;
	  default:
	    ErrMsg("Bad parameter for verbose mode", argv[arg]);
	  }
	} /* loop over characters */
	if (!ROVER_TraceFlag)
	  ROVER_TraceFlag = SHOW_ALL_TRACE;
#endif
	/* endif long list of argument options */
	
      } else {
	ErrMsg("Bad Command Line Argument", argv[arg]);
	usage(argv[0]);
      }
    } else {	   /* If no leading `-' (or ignore is set),
		      then check for main argument */
      if (!keyword) { 	/* Check for successive keyword arguments */
	url = argv[arg];
	if ((AppArgv = (char **)malloc((argc-arg+3)*sizeof(char*))) == NULL) {
	  /* Allocate space for image name and trailing null argv entry */
	  fprintf(stderr, "Out of memory [main:HTBrowse.c]\n");
	  exit(-1);
	}
	AppArgv[keyword++] = url;
      } else {
	if (RoverDebug)
	  fprintf(stderr, "Arg %d is |%s|, storing in %d\n", arg, argv[arg],
		  keyword);
	AppArgv[keyword++] = strdup(argv[arg]);
      }
    } /* Not an option '-'*/
  } /* End of argument loop */

  /* Initialization */
  HTLibInit();
  RoverInit(argc, argv);
  HTEnableFrom = NO;			     /* Send `From:' in the request? */

  /* Open Trace File. */
  if (tracefile_flag) {
    if(!tracefile_root)
      tracefile_root = "/usr/tmp/";
    HTTraceFileName = (char*) malloc(strlen(tracefile_root)+20);

    sprintf(HTTraceFileName, "%s-%d", tracefile_root, (int) getpid());
    HTTracefile = fopen(HTTraceFileName, "a");
    if (!HTTracefile)
      ErrMsg("Can't open trace file", HTTraceFileName);
  };

  if (RoverQRPCTcpTest) {	/* Special check for QRPC Timing tests */
    HTRoverQRPCTest();
    HTLibTerminate();
    Thread_deleteAll();
    return(0);
  } else if (RoverHTTPTestNum) {/* Special check for HTTP Timing tests */
    HTRoverHTTPTest();
    HTLibTerminate();
    Thread_deleteAll();
    return(0);
  }

  state = (QRPCState *)malloc(sizeof(QRPCState));
  state->toFD = -1;
  state->msgID = 0;
  state->type = 0;
  state->session = 0;	/* This will be filled in for us */
  state->url = strdup(url);
  state->obj = NULL;
  state->dv = NULL;
  state->dvsize = 0;
  state->execute = 1;
  state->next = NULL;

  if (!keyword) {
    if ((AppArgv = (char **)malloc((2)*sizeof(char*))) == NULL) {
      /* Allocate space for image name and trailing null argv entry */
      fprintf(stderr, "Out of memory [main:HTBrowse.c]\n");
      RoverExit(-1);
    }
    keyword = 1;
    AppArgv[0] = url;
  }
  AppArgv[keyword] = NULL;

  state->argv = AppArgv;
  state->argc = keyword;
  time(&(state->loadTime));
  if (RoverDebug)
    fprintf(stderr, "Fetching application |%s| w/ %d args\n",url, keyword);
  if (RoverDebug) {
    int i;
    fprintf(stderr, "Args: ");
    for (i = 0; i < keyword; i++)
      fprintf(stderr, "|%s| ", state->argv[i]);
    fprintf(stderr, "\n");
  }

  /* Load and exec the application */
  HTid = App_LocalExec(url, LoadAppCB, state);
  if (HTid < 0) {
    fprintf(stderr, "Application load failed %d [main:HTBrowse.c]\n", HTid);
    RoverExit(-1);
  }

  if ((return_status = HTEventLoopNull()) == HT_ERROR) RoverExit(-1);

  HTLibTerminate();
  Thread_deleteAll();
  return return_status;
}

/* End HTBrowse.c */

