/**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: mosaic.c
 *
 * Description: Mosaic browser interface layer for ROVER
 *
 * Last Modified:  6-July-95 (adj)
 *
 * Global Functions:
 *	void MosaicGotoURL(char* url);
 *	void MosaicEndbrowser();
 *	void MosaicForkbrowser();
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995
 *   Alan deLespinasse and Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pwd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <time.h>


#define ROVER_CLIENT

extern int RoverDebug;		/* Print (lots of) debug info if 1 */

/* Command line for running Mosaic. The first NULL gets filled in just
   before execution-- it specifies the address to use as a proxy server. */
char *mosaiccmd[] = {
#ifdef sparc
  "/home/www/sun4/bin/Mosaic",
#else
  "Mosaic",
#endif /* sparc */
  "-xrm","*defaultWidth: 440","-home","http://home/",
  "","-xrm",NULL,"-xrm","*defaultHeight: 450",NULL,NULL};
int mosaicpid = 0;			/* Mosaic's process ID */
char mrcfile[20];			/* Mosaic's remote control file */

extern void endit(int code);
extern int RoverPort;			/* socket port to user */
extern char LocalHostName[257];		/* Local machine's name */

/* Sends a Mosaic remote-control command. Leaves a file in "/tmp". */
void MosaicGotoURL(char* url)
{
  char *command = "goto";
  /* File should probably be locked. */
  FILE *file = fopen(mrcfile,"w");
  if (file == NULL) {
    perror(mrcfile);
    return;
  }
  fprintf(file ,"%s\n%s\n", command, url ? url : "");
  fclose(file);
  kill(mosaicpid,SIGUSR1);
}


void MosaicEndbrowser()
{
  if (!mosaicpid) return;
  if (kill(mosaicpid,SIGTERM)==0) wait(NULL); /* end Mosaic */
  unlink(mrcfile);		/* erase Mosaic remote control file */
}

void MosaicForkbrowser()
{
  char s[200];
  
  /* Now fork new processes and run Mosaic. The command line must tell Mosaic
     to use the current process as a proxy server. */
  
#ifndef sparc
  unsetenv("WWW_HOME");
#endif /* sparc */

  sprintf(s, "*ProxyGateway: http://%s:%d/", LocalHostName, RoverPort);
  if (RoverDebug) fprintf(stderr, "%s\n", s);
  mosaiccmd[7]=s;
  
  mosaicpid=fork();
  if (mosaicpid == -1) {
    perror("mosaic");
    endit(1);
  }
  if (mosaicpid==0) {
    execvp(mosaiccmd[0],mosaiccmd);
    perror("mosaic");
    endit(1);
  }
  
  sprintf(mrcfile,"/tmp/Mosaic.%d",mosaicpid); /* remote control file */
  if (RoverDebug) fprintf(stderr, "Mosaic rc file: %s\n",mrcfile);
}
